class Callable extends Function {
  constructor(f) {
    super();
    Reflect.setPrototypeOf(f, new.target.prototype);
    return f;
  }

}

class ClassMeta {
  static getMeta(cls) {
    return {
      methods: ClassMeta.getMethodNames(cls),
      staticMethods: ClassMeta.getStaticMethodNames(cls),
      staticProperties: ClassMeta.getStaticPropertyNames(cls)
    };
  }
  /**
   *
   * @param {class|function|object} cls
   * @returns {any|string}
   */


  static getMethodNames(cls) {
    try {
      return !!cls && !!cls.prototype ? Reflect.ownKeys(cls.prototype) : [];
    } catch (e) {
      return 'N/A';
    }
  }
  /**
   *
   * @param {class} cls
   * @return {string[]|string}
   */


  static getStaticMethodNames(cls) {
    try {
      return Reflect.ownKeys(cls).filter(prop => typeof cls[prop] === 'function');
    } catch (e) {
      return 'N/A';
    }
  }
  /**
   *
   * @param {class} cls
   * @return {string[]|string}
   */


  static getStaticPropertyNames(cls) {
    try {
      return Object.keys(cls);
    } catch (e) {
      return 'N/A';
    }
  }

}

class FuncHelper {
  /**
   *
   * @param {function(*): *} pointwiseFunc
   * @return {function(*[]): *[]}
   */
  static rowwise(pointwiseFunc) {
    return row => row.map(x => pointwiseFunc(x));
  }
  /**
   *
   * @param {function(*): *} seinFunc function
   * @param {function(*): *} funcs function whose sole parameter and return value are typed identically
   * @return {function(*): *}
   */


  static chain(seinFunc, ...funcs) {
    return funcs.reduce((curr, prev) => x => prev(curr(x)), seinFunc);
  }
  /**
   *
   * @param {function(*): *} funcs - function whose sole parameter and return value are typed identically
   * @return {function(*): *}
   */


  static pipeline(...funcs) {
    funcs = funcs.filter(x => x);

    switch (funcs.length) {
      case 0:
        return x => x;

      case 1:
        return funcs[0];

      default:
        let f = funcs.shift();
        return funcs.reduce((curr, prev) => x => prev(curr(x)), f);
    } // return funcs.length ? x => funcs.filter(x => x).reduce((a, b) => a |> b, x) : x => x

  }

}

class ProxyFactory {
  /**
   *
   * @param {*} ob
   * @param {ProxyHandler<*>} handler
   * @returns {Proxy<*>}
   */
  static build(ob, handler) {
    return new Proxy(ob, handler);
  }

}

export { Callable, ClassMeta, FuncHelper, ProxyFactory };
